package SecureDistributedKeyGenerationForDiscreteLogBasedCryptosystem;

import Communication.User;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import JointFeldmanProtocol.DistributedKeyGeneration;
import ShamirSecretSharing.Polynomial;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Random;

/**
 * Created by Tzlil on 2/17/2016.
 */
public class SecureDistributedKeyGeneration extends DistributedKeyGeneration {

    private VerifiableSecretSharing verifiableSecretSharing;
    private final BigInteger h;
    private Polynomial.Point[] sharesT;

    public SecureDistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , BigInteger h, Group<BigInteger> group, int id) {
        super(t, n, zi, random, q, g, group, id);
        this.h = h;
        BigInteger r = new BigInteger(q.bitLength(),random).mod(q);
        this.verifiableSecretSharing = new VerifiableSecretSharing(t,n,r,random,q,h,group);
    }

    @Override
    public void sendSecret(User user,int j) {
        Polynomial.Point secret = getShare(j);
        Polynomial.Point secretT = verifiableSecretSharing.getShare(j);
        DKGMessages.DoubleSecretMessage doubleSecretMessage = doubleSecretMessage(id,j,secret,secretT);
        user.send(j, DKGMessages.Mail.Type.SECRET, doubleSecretMessage);
    }

    @Override
    public boolean isValidSecret(int i, BigInteger[] commitments, int j){
        Polynomial.Point secret = shares[i - 1];
        Polynomial.Point secretT = sharesT[i - 1];
        return isValidSecret(secret,secretT,commitments, j);
    }

    public boolean isValidSecret(Polynomial.Point secret,Polynomial.Point secretT, BigInteger[] verificationValues, int j){
        BigInteger v = verify(j,verificationValues,group);
        BigInteger exp = group.add(group.multiply(g, secret.y),group.multiply(h, secretT.y));
        return exp.equals(v);
    }


    public void broadcastComplaint(User user,Polynomial.Point secret,Polynomial.Point secretT,int i){
        DKGMessages.DoubleSecretMessage complaint = doubleSecretMessage(i,id,secret,secretT);
        user.broadcast(DKGMessages.Mail.Type.COMPLAINT,complaint);
    }

    public void broadcastAnswer(User user,Polynomial.Point secret,Polynomial.Point secretT,int i){
        DKGMessages.DoubleSecretMessage complaint = doubleSecretMessage(i,id,secret,secretT);
        user.broadcast(DKGMessages.Mail.Type.ANSWER,complaint);
    }

    /**
     * stage4.3 according to the protocol
     *  if check fails for index i, Pj
     */
    public void broadcastComplaints(User user, BigInteger[][] commitmentsTable, boolean stage4){
        if(!stage4){
            broadcastComplains(user,commitmentsTable);
        }else{
            for (int i = 1; i <= n ; i++ ){
                if(i != id) {
                    if (!isValidSecret(shares[i - 1],commitmentsTable[i - 1],id)) {
                        broadcastComplaint(user,shares[i - 1],sharesT[i - 1],i);
                    }
                }
            }
        }
    }


    public void broadcastVerificationValues(User user){
        BigInteger[] verificationValues = new BigInteger[t + 1];
        BigInteger[] hBaseCommitments = verifiableSecretSharing.getCommitmentsArray();
        for (int k = 0 ; k < verificationValues.length ; k++){
            verificationValues[k] = group.add(commitmentsArray[k],hBaseCommitments[k]);
        }
        broadcastCommitments(user,verificationValues);
    }

    private DKGMessages.DoubleSecretMessage doubleSecretMessage(int i,int j,Polynomial.Point secret, Polynomial.Point secretT){
        DKGMessages.DoubleSecretMessage doubleSecretMessage = DKGMessages.DoubleSecretMessage.newBuilder()
                .setI(i)
                .setJ(j)
                .setSecret(ByteString.copyFrom(secret.y.toByteArray()))
                .setSecretT(ByteString.copyFrom(secretT.y.toByteArray()))
                .build();
        return doubleSecretMessage;
    }

    @Override
    public void broadcastComplaintAnswer(User user, int j) {
        DKGMessages.DoubleSecretMessage answer = doubleSecretMessage(id,j,getShare(j)
                ,verifiableSecretSharing.getShare(j));
        user.broadcast(DKGMessages.Mail.Type.ANSWER,answer);
    }

    public void setSharesT(Polynomial.Point[] sharesT) {
        this.sharesT = sharesT;
    }
}
